!-----------------------------------------------------------------------------
!> A prototype solver for advection-diffusion-reaction equation,
!> This equation is generic and intended for education purposes
!> but may also serve as a starting point for more complex solvers.
!------------------------------------------------------------------------------
SUBROUTINE AdvDiffSolver_Fluxcompute( Model,Solver,dt,TransientSimulation )
!------------------------------------------------------------------------------
  USE DefUtils

  IMPLICIT NONE
!------------------------------------------------------------------------------
  TYPE(Solver_t) :: Solver
  TYPE(Model_t) :: Model
  REAL(KIND=dp) :: dt
  LOGICAL :: TransientSimulation
!------------------------------------------------------------------------------
! Local variables
!------------------------------------------------------------------------------
  TYPE(Element_t),POINTER :: Element
  REAL(KIND=dp) :: Norm
  INTEGER :: n, nb, nd, t, active
  INTEGER :: iter, maxiter
  LOGICAL :: Found
!------------------------------------------------------------------------------

  CALL DefaultStart()
  
  maxiter = ListGetInteger( GetSolverParams(),&
      'Nonlinear System Max Iterations',Found,minv=1)
  IF(.NOT. Found ) maxiter = 1

  ! Nonlinear iteration loop:
  !--------------------------
  DO iter=1,maxiter

    ! System assembly:
    !----------------
    CALL DefaultInitialize()
    Active = GetNOFActive()
    DO t=1,Active
      Element => GetActiveElement(t)
      n  = GetElementNOFNodes()
      nd = GetElementNOFDOFs()
      nb = GetElementNOFBDOFs()
      CALL LocalMatrix(  Element, n, nd+nb )
    END DO

    CALL DefaultFinishBulkAssembly()

    Active = GetNOFBoundaryElements()
    DO t=1,Active
      Element => GetBoundaryElement(t)
      IF(ActiveBoundaryElement()) THEN
        n  = GetElementNOFNodes()
        nd = GetElementNOFDOFs()
        nb = GetElementNOFBDOFs()
        CALL LocalMatrixBC(  Element, n, nd+nb )
      END IF
    END DO

    CALL DefaultFinishBoundaryAssembly()
    CALL DefaultFinishAssembly()
    CALL DefaultDirichletBCs()

    ! And finally, solve:
    !--------------------
    Norm = DefaultSolve()

    IF( Solver % Variable % NonlinConverged > 0 ) EXIT

  END DO

  CALL DefaultFinish()
  
CONTAINS

! Assembly of the matrix entries arising from the bulk elements
!------------------------------------------------------------------------------
  SUBROUTINE LocalMatrix( Element, n, nd )
!------------------------------------------------------------------------------
    INTEGER :: n, nd
    TYPE(Element_t), POINTER :: Element
!------------------------------------------------------------------------------
    REAL(KIND=dp) :: diff_coeff(n), conv_coeff(n),react_coeff(n), &
                     time_coeff(n), D,C,R, rho,Velo(3,n),a(3), Weight
    REAL(KIND=dp) :: Basis(nd),dBasisdx(nd,3),DetJ,LoadAtIP
    REAL(KIND=dp) :: MASS(nd,nd), STIFF(nd,nd), FORCE(nd), LOAD(n)
    LOGICAL :: Stat,Found
    INTEGER :: i,t,p,q,dim
    TYPE(GaussIntegrationPoints_t) :: IP
    TYPE(ValueList_t), POINTER :: BodyForce, Material
    TYPE(Nodes_t) :: Nodes
    ! added variables
    TYPE(Variable_t), POINTER :: VarPressure
    REAL(KIND=dp), POINTER :: Pressure(:), gWork(:,:) ! needed to receive gravity
    INTEGER, POINTER :: PressurePerm(:)
    INTEGER :: component
    REAL(KIND=dp) :: HydrCond(n), HydrCondAtIP, gradPAtIP(3), FluxAtIP(3),Gravity(3),waterdens(n),waterdensAtIP
    
    SAVE Nodes
!------------------------------------------------------------------------------

    dim = CoordinateSystemDimension()
    
    VarPressure => VariableGet( Model % Mesh % Variables, TRIM('GWPressure'))
    IF (.NOT.ASSOCIATED(VarPressure)) THEN
      CALL FATAL( "ModelPDE_Temperature", ' "GWPressure" not found')
    END IF
    PressurePerm => VarPressure % Perm
    Pressure => VarPressure % Values

    component = GetInteger(GetSolverParams(), 'component', Found)
    if (.NOT.Found) CALL FATAL("ModelPDE_Fluxcompute","component not found")
    
    gWork => ListGetConstRealArray( Model % Constants,'Gravity',Found)
    IF (.NOT.Found) THEN
      Gravity = 0.0
      CALL WARN("ModelPDE_Darcy",'Gravity not found in Constants section. Setting to zero')
    ELSE
      !PRINT *,gWork(1:4,1)
      Gravity = gWork(1:3,1)*gWork(4,1)
      !PRINT *,"g=(",Gravity(1:DIM),")"
    END IF    

    CALL GetElementNodes( Nodes )
    MASS  = 0._dp
    STIFF = 0._dp
    FORCE = 0._dp
    LOAD = 0._dp

    BodyForce => GetBodyForce()
    IF ( ASSOCIATED(BodyForce) ) &
       Load(1:n) = GetReal( BodyForce,'Heat source', Found )

    Material => GetMaterial()
    HydrCond(1:n)=GetReal(Material,'Hydrological Conductivity',Found)
    !diff_coeff(1:n)=GetReal(Material,'Eff Heat Conductivity',Found)
    !react_coeff(1:n)=GetReal(Material,'reaction coefficient',Found)
    react_coeff(1:n)=1.0_dp
    !conv_coeff(1:n)=GetReal(Material,'convection coefficient',Found)
    !time_coeff(1:n)=GetReal(Material,'time derivative coefficient',Found)
    !time_coeff(1:n)= 1.0_dp ! set to have unit mass matrix
    waterdens(1:n)=GetReal(Material,'Water Density',Found)
    
    !Velo = 0._dp
    !DO i=1,dim
    !  Velo(i,1:n)=GetReal(Material,&
    !      'convection velocity '//TRIM(I2S(i)),Found)
    !END DO

    ! Numerical integration:
    !-----------------------
    IP = GaussPoints( Element )
    DO t=1,IP % n
      ! Basis function values & derivatives at the integration point:
      !--------------------------------------------------------------
      stat = ElementInfo( Element, Nodes, IP % U(t), IP % V(t), &
           IP % W(t), detJ, Basis, dBasisdx )
      !gradPAtIP = 0.0_dp
     ! DO I=1,DIM
     !   gradPAtIP(I) = SUM(Pressure(PressurePerm(Element % NodeIndexes(1:n))) &
     !          * dBasisdx(1:n,I))
     ! END DO

      gradPAtIP = MATMUL(Pressure(PressurePerm(Element % NodeIndexes(1:n))),dBasisdx(1:n,:))      
      HydrCondAtIP = SUM(HydrCond(1:n)*Basis(1:n))
      waterdensAtIP = SUM(waterdens(1:n)*Basis(1:n))
      fluxAtIP(1:dim) = HydrCondAtIP * (waterdensAtIP*Gravity(1:dim) - gradPAtIP(1:dim))
      
      ! The source term at the integration point:
      !------------------------------------------
      LoadAtIP = SUM( Basis(1:n) * LOAD(1:n) )

      rho = SUM(Basis(1:n)*time_coeff(1:n))
      !a = MATMUL(Velo(:,1:n),Basis(1:n))
      !D = SUM(Basis(1:n)*diff_coeff(1:n))
      !C = SUM(Basis(1:n)*conv_coeff(1:n))
      R = SUM(Basis(1:n)*react_coeff(1:n))

      Weight = IP % s(t) * DetJ

      ! diffusion term (D*grad(u),grad(v)):
      ! -----------------------------------
      !STIFF(1:nd,1:nd) = STIFF(1:nd,1:nd) + Weight * &
      !       D * MATMUL( dBasisdx, TRANSPOSE( dBasisdx ) )

      DO p=1,nd
        DO q=1,nd
          ! advection term (C*grad(u),v)
          ! -----------------------------------
          !STIFF (p,q) = STIFF(p,q) + Weight * &
          !   C * SUM(a(1:dim)*dBasisdx(q,1:dim)) * Basis(p)
          !STIFF (p,q) = STIFF(p,q) + Weight *&
          !     C* SUM(FluxAtIP(1:dim)*dBasisdx(q,1:dim)) * Basis(p)

          ! reaction term (R*u,v)
          ! -----------------------------------
          STIFF(p,q) = STIFF(p,q) + Weight * R*Basis(q) * Basis(p)

          ! time derivative (rho*du/dt,v):
          ! ------------------------------
          !MASS(p,q) = MASS(p,q) + Weight  * Basis(q) * Basis(p)
        END DO
      END DO
      LoadAtIP = fluxAtIP(component)
      FORCE(1:nd) = FORCE(1:nd) + Weight * LoadAtIP * Basis(1:nd)
    END DO

    IF(TransientSimulation) CALL Default1stOrderTime(MASS,STIFF,FORCE)
    CALL LCondensate( nd-nb, nb, STIFF, FORCE )
    CALL DefaultUpdateEquations(STIFF,FORCE)
!------------------------------------------------------------------------------
  END SUBROUTINE LocalMatrix
!------------------------------------------------------------------------------


! Assembly of the matrix entries arising from the Neumann and Robin conditions
!------------------------------------------------------------------------------
  SUBROUTINE LocalMatrixBC( Element, n, nd )
!------------------------------------------------------------------------------
    INTEGER :: n, nd
    TYPE(Element_t), POINTER :: Element
!------------------------------------------------------------------------------
    REAL(KIND=dp) :: Flux(n), Coeff(n), Ext_t(n), F,C,Ext, Weight
    REAL(KIND=dp) :: Basis(nd),dBasisdx(nd,3),DetJ,LoadAtIP
    REAL(KIND=dp) :: STIFF(nd,nd), FORCE(nd), LOAD(n)
    LOGICAL :: Stat,Found
    INTEGER :: i,t,p,q,dim
    TYPE(GaussIntegrationPoints_t) :: IP

    TYPE(ValueList_t), POINTER :: BC

    TYPE(Nodes_t) :: Nodes
    SAVE Nodes
!------------------------------------------------------------------------------
    BC => GetBC()
    IF (.NOT.ASSOCIATED(BC) ) RETURN

    dim = CoordinateSystemDimension()

    CALL GetElementNodes( Nodes )
    STIFF = 0._dp
    FORCE = 0._dp
    LOAD = 0._dp

    ! flux keyword changed
    Flux(1:n)  = GetReal( BC,'Groundwater flux', Found )
    !Coeff(1:n) = GetReal( BC,'robin coefficient', Found )
    !Ext_t(1:n) = GetReal( BC,'external field', Found )
    Coeff(1:n) = 0.0_dp
    Ext_t(1:n) = 0.0_dp

    ! Numerical integration:
    !-----------------------
    IP = GaussPoints( Element )
    DO t=1,IP % n
      ! Basis function values & derivatives at the integration point:
      !--------------------------------------------------------------
      stat = ElementInfo( Element, Nodes, IP % U(t), IP % V(t), &
              IP % W(t), detJ, Basis, dBasisdx )

      Weight = IP % s(t) * DetJ

      ! Evaluate terms at the integration point:
      !------------------------------------------

      ! Given flux:
      ! -----------
      F = SUM(Basis(1:n)*flux(1:n))

      ! Robin condition (C*(u-u_0)):
      ! ---------------------------
      C = SUM(Basis(1:n)*coeff(1:n))
      Ext = SUM(Basis(1:n)*ext_t(1:n))

      DO p=1,nd
        DO q=1,nd
          STIFF(p,q) = STIFF(p,q) + Weight * C * Basis(q) * Basis(p)
        END DO
      END DO

      FORCE(1:nd) = FORCE(1:nd) + Weight * (F + C*Ext) * Basis(1:nd)
    END DO
    CALL DefaultUpdateEquations(STIFF,FORCE)
!------------------------------------------------------------------------------
  END SUBROUTINE LocalMatrixBC
!------------------------------------------------------------------------------

! Perform static condensation in case bubble dofs are present
!------------------------------------------------------------------------------
  SUBROUTINE LCondensate( N, Nb, K, F )
!------------------------------------------------------------------------------
    USE LinearAlgebra
    INTEGER :: N, Nb
    REAL(KIND=dp) :: K(:,:),F(:),Kbb(Nb,Nb), &
         Kbl(Nb,N), Klb(N,Nb), Fb(Nb)

    INTEGER :: m, i, j, l, p, Ldofs(N), Bdofs(Nb)

    IF ( Nb <= 0 ) RETURN

    Ldofs = (/ (i, i=1,n) /)
    Bdofs = (/ (i, i=n+1,n+nb) /)

    Kbb = K(Bdofs,Bdofs)
    Kbl = K(Bdofs,Ldofs)
    Klb = K(Ldofs,Bdofs)
    Fb  = F(Bdofs)

    CALL InvertMatrix( Kbb,nb )

    F(1:n) = F(1:n) - MATMUL( Klb, MATMUL( Kbb, Fb  ) )
    K(1:n,1:n) = K(1:n,1:n) - MATMUL( Klb, MATMUL( Kbb, Kbl ) )
!------------------------------------------------------------------------------
  END SUBROUTINE LCondensate
!------------------------------------------------------------------------------

!------------------------------------------------------------------------------
END SUBROUTINE AdvDiffSolver_Fluxcompute
!------------------------------------------------------------------------------
